%{

/*
   cfg.l -- PIKT piktc lexer for all .cfg files except systems.cfg

   $Header: /var/cvs/PIKT/pikt/src/piktc/cfg.l,v 1.82 2007/01/02 09:04:17 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2006 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org

   ----------------------------------------------------------------------------

   this is the sausage factory; not for the squeamish; enter at your own risk!

   on our test irix system, flex failed to match regexps of the form
   '{D}{x,y}', which led to parser failures; hence substitutions of the
   form '{D}+' below

 */

#include "../lib/common.h"
#include "piktc.h"
#include "cfg.tab.h"

char _cfglex_[] = "cfglex";

#define RETURN(token) cfglval.sval = mystrdup(yytext, _cfglex_); return token

#undef YY_INPUT
#define YY_INPUT(buf, result, max_size) { \
		signed char c = cfginput(); \
		result = (c == EOF) ? YY_NULL : (buf[0] = c, 1); \
	}

int ii, jj;
int fn = 0;

cfgln *clpcurr = (cfgln *) NULL;

char varnum[(MAXIDENT+1)*MAXVARS+2];	/* this is the appropriate size if we
					   check to avoid duplicate entries;
					   otherwise, this is arbitrarily
					   large, but plenty large, practically
					   speaking
					 */
char varstr[(MAXIDENT+1)*MAXVARS+2];	/* ditto */

void cfgerror(const char *);
signed int cfginput(void);
int regvar(char *, int);
int vardattype(char *, int, cfgln *);

%}

W		[ \t]
nW		[^ \t]
WN		[ \t\n]
nWN		[^ \t\n]
WND		[ \t\n0-9]
D		[0-9]
L		[A-Za-z]
DL		[0-9A-Za-z]
DLU		[0-9A-Za-z_]
nB		[^\[\]]

%x MDS MPS AS ASE PS FS OS CS DS SS FNS LS CMNT2

%option noyywrap

%%

"/"{3,8}\n			{ ii = 0;
				  switch(strlen(yytext)-1) {
					case 3:
						if (isverbose) {
							if (islist_elements) {
								printf("listing macro elements...\n");
								fflush(stdout);
							}
						}
						return _MCRCFGBOF;
						break;
					case 4:
						if (isverbose) {
							if (islist_elements) {
								printf("listing alarm elements...\n");
								fflush(stdout);
							}
						}
						return _ALMCFGBOF;
						break;
					case 5:
						if (isverbose) {
							if (islist_elements) {
								printf("listing alert elements...\n");
								fflush(stdout);
							}
						}
						return _ALTCFGBOF;
						break;
					case 6:
						if (isverbose) {
							if (islist_elements) {
								printf("listing program elements...\n");
								fflush(stdout);
							}
						}
						return _PRGCFGBOF;
						break;
					case 7:
						if (isverbose) {
							if (islist_elements) {
								printf("listing file elements...\n");
								fflush(stdout);
							}
						}
						return _FILCFGBOF;
						break;
					case 8:
						if (isverbose) {
							if (islist_elements) {
								printf("listing object elements...\n");
								fflush(stdout);
							}
						}
						return _OBJCFGBOF;
						break;
				  }
				}

<MDS,AS,LS,OS>"/"{3,8}\n	{ yyless(0);
				  BEGIN(INITIAL);
				}

<MDS,AS,LS,OS>\177"/"+\n	{ /* <MDS,AS,LS,OS>\177"/"{3,8}\n */
				  yyless(1);
				  BEGIN(INITIAL);
				}

<*>^{D}+{W}*\n			{ lineno = atoi(yytext); }

<*>\177				{ ; /* dummy char; see
				       cfginput() below
				     */
				}

<*>^({L}|"/")[-_\./0-9A-Za-z]*	{ if ((yytext[0] == '/') && (cfgcurr != PRGCFG) && (cfgcurr != FILCFG))
					cfgerror("syntax error");
				  if (yyleng > MAXIDENT) {
					logreturn(OUTPUT, LOG_ERR, ERRID_IDENT_MAX_LENGTH_EXCEEDED, __FILE__, __LINE__, _cfglex_, FALSE, "stanza name length exceeds MAXIDENT (%d): %s", yytext, MAXIDENT);
					bailout = TRUE;
					*(yytext+MAXIDENT) = '\0';
				  }
				  switch(cfgcurr) {
				  	case MCRCFG:
						BEGIN(MDS);
						break;
					case ALMCFG:
						varnum[0] = ' '; varnum[1] = '\0';
						varstr[0] = ' '; varstr[1] = '\0';
						BEGIN(INITIAL);
						break;
					case ALTCFG:
						BEGIN(AS);
						break;
					case PRGCFG:
						ii = 0;
						BEGIN(PS);
						break;
					case FILCFG:
						ii = 0;
						BEGIN(FS);
						break;
					case OBJCFG:
						BEGIN(OS);
						break;
					default:
					  	BEGIN(INITIAL);
						break;
				  }
				  if (islist_elements) {
					printf("%s\n", yytext);
					fflush(stdout);
				  }
				  RETURN(_STANZA);
				}

<INITIAL,MDS>^{L}[-_\.0-9A-Za-z]*/"("{nWN}+ {
				  if (cfgcurr == MCRCFG) {
					if (yyleng > MAXIDENT) {
						logreturn(OUTPUT, LOG_ERR, ERRID_IDENT_MAX_LENGTH_EXCEEDED, __FILE__, __LINE__, _cfglex_, FALSE, "stanza name length exceeds MAXIDENT (%d): %s", yytext, MAXIDENT);
						bailout = TRUE;
						*(yytext+MAXIDENT) = '\0';
					}
					BEGIN(MPS);
					if (islist_elements) {
						printf("%s\n", yytext);
						fflush(stdout);
					}
					RETURN(_STANZA);
				  } else
					cfgerror("syntax error");
				}

<MDS>=(piktnullchar|_)		{ cfglval.sval = mystrdup(piktnullchar, _cfglex_);
				  return _MACRODEF;
				}
<MDS>=piktspacechar		{ cfglval.sval = mystrdup(piktspacechar, _cfglex_);
				  return _MACRODEF;
				}
<MDS>=pikttabchar		{ cfglval.sval = mystrdup(pikttabchar, _cfglex_);
				  return _MACRODEF;
				}
<MDS>=piktnewlinechar		{ cfglval.sval = mystrdup(piktnewlinechar, _cfglex_);
				  return _MACRODEF;
				}
<MDS>{nWN}+			{ RETURN(_MACRODEF); }
<MDS>\"[^\"]*\"			{ RETURN(_MACRODEF); }

<MPS>"("			{ return yytext[0]; }
<MPS>")"			{ BEGIN(MDS);
				  return yytext[0];
				}
<MPS>","			{ return yytext[0]; }
<MPS>{L}{DLU}*			{ RETURN(_MACROPRM); }

<INITIAL,AS,DS,SS>status	{ RETURN(_STATUS); }
<INITIAL,AS,DS,SS>active	{ RETURN(_ACTIVE); }
<INITIAL,AS,DS,SS>inactive	{ RETURN(_INACTIVE); }
<INITIAL,AS,DS,SS>suspended	{ RETURN(_SUSPENDED); }
<INITIAL,AS,DS,SS>testing	{ RETURN(_TESTING); }
<INITIAL,AS,DS,SS>debugging	{ RETURN(_DEBUGGING); }

<INITIAL,AS,DS,SS>level		{ RETURN(_LEVEL); }
<INITIAL,AS,DS,SS>emerg|emergency { RETURN(_EMERG); }
<INITIAL,AS,DS,SS>alert|urgent	{ RETURN(_ALERT); }
<INITIAL,AS,DS,SS>crit|critical	{ RETURN(_CRIT); }
<INITIAL,AS,DS,SS>err|error	{ RETURN(_ERR); }
<INITIAL,AS,DS,SS>warn|warning	{ RETURN(_WARNING); }
<INITIAL,AS,DS,SS>notice	{ RETURN(_NOTICE); }
<INITIAL,AS,DS,SS>info		{ RETURN(_INFO); }
<INITIAL,AS,DS,SS>debug		{ RETURN(_DEBUG); }

<INITIAL,AS>__PIKTSTATUS__	{ RETURN(_PIKTSTATUS); }
<INITIAL,AS>__PIKTLEVEL__	{ RETURN(_PIKTLEVEL); }

<INITIAL,DS,SS>task		{ RETURN(_TASK); }

<INITIAL,DS,SS>input		{ RETURN(_INPUT); }
<INITIAL,DS,SS>proc|process	{ BEGIN(CS);
				  RETURN(_PROCESS);
				}
<INITIAL,DS,SS>file		{ RETURN(_REGFILE); }
<INITIAL,DS,SS>log|logfile	{ RETURN(_LOGFILE); }

<INITIAL,DS,SS>filter		{ BEGIN(CS);
				  RETURN(_FILTER);
				}

<INITIAL,DS,SS>seps|separators	{ RETURN(_SEPARATORS); }
<INITIAL,DS,SS>dat|data		{ BEGIN(DS);
				  RETURN(_DAT);
				}
<INITIAL,DS,SS>keys		{ BEGIN(DS);
				  RETURN(_KEYS);
				}

<INITIAL,DS,SS>init		{ RETURN(_INIT); }
<INITIAL,DS,SS,CS>begin		{ BEGIN(INITIAL);
				  RETURN(_BEGIN);
				}
<INITIAL,DS,SS,CS>end		{ BEGIN(INITIAL);
				  RETURN(_END);
				}
<INITIAL,DS,SS,CS>rule		{ BEGIN(INITIAL);
				  RETURN(_RULE);
				}

<INITIAL,DS,SS>do		{ RETURN(_DO); }
<INITIAL,DS,SS>if		{ RETURN(_IF); }
<INITIAL,DS,SS>if{WND}*"("	{ unput('[');
				  cfglval.sval = mystrdup("if", _cfglex_);
				  return _IF;
				}
<INITIAL,DS,SS>(elif|elsif|elseif) { RETURN(_ELIF); }
<INITIAL,DS,SS>(elif|elsif|elseif){WND}*"(" {
				  unput('[');
				  cfglval.sval = mystrdup("elif", _cfglex_);
				  return _ELIF;
				}
<INITIAL,DS,SS>else		{ RETURN(_ELSE); }
<INITIAL,DS,SS>endif|fi		{ RETURN(_ENDIF); }
<INITIAL,DS,SS>while		{ RETURN(_WHILE); }
<INITIAL,DS,SS>while{WND}*"("	{ unput('[');
				  cfglval.sval = mystrdup("while", _cfglex_);
				  return _WHILE;
				}
<INITIAL,DS,SS>endwhile		{ RETURN(_ENDWHILE); }
<INITIAL,DS,SS>repeat		{ RETURN(_REPEAT); }
<INITIAL,DS,SS>until		{ RETURN(_UNTIL); }
<INITIAL,DS,SS>until{WND}*"("	{ unput('[');
				  cfglval.sval = mystrdup("until", _cfglex_);
				  return _UNTIL;
				}
<INITIAL,DS,SS>for		{ RETURN(_FOR);	}
<INITIAL,DS,SS>for{WND}*"("	{ RETURN(_FORLP); }
<INITIAL,DS,SS>foreach		{ RETURN(_FOREACH); }
<INITIAL,DS,SS>endfor|endforeach { RETURN(_ENDFOR); }
<INITIAL,DS,SS>in		{ RETURN(_IN); }
<INITIAL,DS,SS>break		{ RETURN(_BREAK); }
<INITIAL,DS,SS>continue|cont	{ RETURN(_CONTINUE); }
<INITIAL,DS,SS>switch		{ RETURN(_SWITCH); }
<INITIAL,DS,SS>switch{WND}*"("	{ RETURN(_SWITCHLP); }
<INITIAL,DS,SS>case		{ RETURN(_CASE); }
<INITIAL,DS,SS>default		{ RETURN(_DEFAULT); }
<INITIAL,DS,SS>breakswitch|breaksw { RETURN(_BREAKSWITCH); }
<INITIAL,DS,SS>endswitch|endsw	{ RETURN(_ENDSWITCH); }
<INITIAL,DS,SS>again		{ RETURN(_AGAIN); }
<INITIAL,DS,SS>leave		{ RETURN(_LEAVE); }
<INITIAL,DS,SS>redo		{ RETURN(_REDO); }
<INITIAL,DS,SS>next		{ RETURN(_NEXT); }
<INITIAL,DS,SS>skip		{ RETURN(_SKIP); }
<INITIAL,DS,SS>last		{ RETURN(_LAST); }
<INITIAL,DS,SS>pause		{ RETURN(_PAUSE); }
<INITIAL,DS,SS>quit		{ RETURN(_QUIT); }
<INITIAL,DS,SS>die		{ RETURN(_DIE);	}

<INITIAL,DS,SS>output		{ RETURN(_OUTPUT); }
<INITIAL,DS,SS>mail		{ RETURN(_MAIL); }
<INITIAL,DS,SS>print		{ RETURN(_PRINT); }
<INITIAL,DS,SS>piktlog		{ RETURN(_PIKTLOG); }
<INITIAL,DS,SS>syslog		{ RETURN(_SYSLOG); }

<INITIAL,DS,SS>set		{ BEGIN(SS);
				  RETURN(_SET);
				}
<INITIAL,DS,SS>unset|delete	{ BEGIN(SS);
				  RETURN(_UNSET);
				}

<INITIAL,DS,SS>exec		{ BEGIN(INITIAL);
				  RETURN(_EXEC);
				}
<INITIAL>wait			{ RETURN(_WAIT); }

<CS,DS>\"[^\n]+\"/{W}*\n	{ BEGIN(INITIAL);
				  RETURN(_STR);
				}
<DS>\"[^\"]+\"			{ RETURN(_STR); }

<DS>({D}+|$|$-{D}+)/[^\]0-9A-Za-z_] {
				  /* <DS>({D}{1,5}|$|$-{D}{1,5})/[^\]0-9A-Za-z_] */
				  BEGIN(INITIAL);
				  RETURN(_POSITION);
				}
<DS>"["({D}+|$|$-{D}+)"]" 	{ /* <DS>"["({D}{1,5}|$|$-{D}{1,5})"]" */
				  BEGIN(INITIAL);
				  RETURN(_POSITION);
				}
<DS>{D}+","({D}+|$)		{ BEGIN(INITIAL);
				  RETURN(_RANGE);
				}
<DS>"["{D}+","({D}+|$)"]"	{ BEGIN(INITIAL);
				  RETURN(_RANGE);
				}

<DS>(#|$){L}{DLU}*"[" 		{ yytext[yyleng-1] = '\0';
				  regvar(yytext+1, (yytext[0] == '#') ? NUMBER : STRING);
				  cfglval.sval = mystrdup(yytext, _cfglex_);
				  unput('[');
				  unput('[');
				  unput('[');
				  return (*cfglval.sval == '#' ? _VARNUM : _VARSTR);
				}

<INITIAL,SS>"="			{ BEGIN(INITIAL);
				  return yytext[0];
				}

<INITIAL,SS>("+="|"-="|"*="|"/="|"%="|"**="|"<<="|">>="|"&="|"|="|"^=") {
				  BEGIN(INITIAL);
				  RETURN(_AOPNUM);
				}
<INITIAL,SS>".="		{ BEGIN(INITIAL);
				  RETURN(_AOPSTR);
				}

<FNS>[A-Z][0-9A-Z]*		{ RETURN(_FLHNDL); }

<INITIAL,SS,FNS>${D}+		{ /* <INITIAL,SS,FNS>${D}{1,5} */
				  RETURN(_VALSTR);
				}

<INITIAL,DS,SS,FNS>(#|$){L}{DLU}* {
				  regvar(yytext+1, (yytext[0] == '#') ? NUMBER : STRING);
				  RETURN(yytext[0] == '#' ? _VARNUM : _VARSTR);
				}

<INITIAL,SS,FNS>$"["		{ yytext[yyleng-1] = '\0';
				  cfglval.sval = mystrdup(yytext, _cfglex_);
				  unput('[');
				  unput('[');
				  unput('[');
				  return _VALSTR;
				}

<INITIAL,SS,FS>(#|$){L}{DLU}*"[" {
				  yytext[yyleng-1] = '\0';
				  regvar(yytext+1, (yytext[0] == '#') ? NUMBER : STRING);
				  cfglval.sval = mystrdup(yytext, _cfglex_);
				  unput('[');
				  unput('[');
				  unput('[');
				  return (*cfglval.sval == '#' ? _VARNUM : _VARSTR);
				}

<INITIAL,FNS>(%|@){L}{DLU}*	{ switch(vardattype(yytext+1, yyleng-1, clpcurr)) {
					case NUMBER:
						RETURN(yytext[0] == '%' ? _VARHSTNUM : _VARPRCNUM);
						break;
					case STRING:
						RETURN(yytext[0] == '%' ? _VARHSTSTR : _VARPRCSTR);
						break;
					case ERROR:
						logreturn(OUTPUT, LOG_ERR, ERRID_DATTYPE_INDETERMINATE, __FILE__, __LINE__, _cfglex_, FALSE, "unable to determine dattype for %s", yytext);
						bailout = TRUE;
						break;
				  }
				}

<INITIAL,DS,SS,FS>"["/{WN}*($|\") { return yytext[0]; }
<INITIAL,DS,SS,FNS>"["		{ return yytext[0]; }
<INITIAL,DS,SS,FNS>"[[["	{ return '['; }
<INITIAL,DS,SS,FNS>{D}+		{ /* <INITIAL,DS,SS,FNS>{D}{1,5} */
				  RETURN(_INT);
				}
<INITIAL,DS,SS,FNS>"]"		{ return yytext[0]; }

<INITIAL,SS,FNS>("+"|"-"|"~")	{ return yytext[0]; }

<INITIAL,SS,FNS>("*"|"/"|"%"|"**"|"&"|"|"|"^"|"<<"|">>") {
				  RETURN(_NOP);
				}

<INITIAL,SS,FNS>"."		{ RETURN(_SOP); }

<INITIAL,SS,FNS>(-e|-z|-f|-d|-c|-b|-p|-l|-S) {
				  RETURN(_FOP);
				}

<INITIAL,SS,FNS>("=="|">"|"<"|">="|"<="|"!>"|"!>="|"!<"|"!<="|"!="|"<>") {
				  RETURN(_NREL);
				}
<INITIAL,SS,FNS>("=~"|"=~~"|"!~"|"!~~"|eq|ne|lt|gt|le|ge) {
				  RETURN(_SREL);
				}

<INITIAL,SS,FNS>(({D}+)|({D}*\.{D}+)|((({D}+)|({D}*\.{D}+))\/(({D}+)|({D}*\.{D}+))))%* {
				  RETURN(_NUM);
				}

<INITIAL,SS,FNS>(#|$){L}{DLU}*/"(" {
				  BEGIN(FNS);
				  RETURN(yytext[0] == '#' ? _FNNUM : _FNSTR);
				}
<SS,FNS>"("			{ fn++;
				  return '(';
				}
<SS,FNS>")"			{ fn--;
				  if (! fn)
					BEGIN(INITIAL);
				  return ')';
				}

<INITIAL,SS,FS,FNS>\"[^\n\"]*\"	{ jj = leadstr(yytext);
				  for (ii=yyleng-1; ii>=jj; ii--)
					unput(*(yytext+ii));
				  yytext[ii+1] = '\0';
				  RETURN(_STR);
				}

<INITIAL>"("{WN}*{nWN}+{WND}*(=~|=~~|!~|!~~|eq|ne|lt|gt|le|ge){WND}*{nWN}+{WN}*")"{WN}* { 
				  yyless(1);
				  return '[';
				}

<INITIAL>"("([-+*/%$#@~^><)(,.0-9A-Za-z_\n\t\[\] ]|![^=~><]|"&"[^&]|"|"[^|])+ {
				  if (condparen(yytext)) {
					yyless(1);
					return '[';
				  } else {
					yyless(1);
					return yytext[0];
				  }
				}

<INITIAL>")"				{ return yytext[0]; }

<INITIAL,SS,FNS>"!"		{ RETURN(_NOT); }
<INITIAL,SS,FNS>"&&"		{ RETURN(_AND);	}
<INITIAL,SS,FNS>"||"		{ RETURN(_OR); }

<AS>timing			{ return _TIMING; }
<AS>drift			{ return _DRIFT; }
<AS>(priority|nice)		{ return _PRIORITY; }
<AS>nicecmd			{ return _NICECMD; }
<AS>mailcmd			{ return _MAILCMD; }
<AS>lpcmd			{ return _LPCMD; }
<AS>execcmd			{ BEGIN(ASE);
				  return _EXECCMD;
				}
<AS>(alarms|scripts)		{ return _ALARMS; }
<AS>[-,*0-9]+			{ RETURN(_TIMEVAL); }
<AS>{D}+-{D}+			{ /* <AS>{D}{1,2}-{D}{1,2} */
				  RETURN(_TIMEVAL); }
<AS>"*/"{D}+			{ /* <AS>"\*\/"{D}{1,2} */
				  RETURN(_TIMEVAL); }
<AS>{D}+-{D}+"/"{D}+		{ /* <AS>{D}{1,2}-{D}{1,2}"/"{D}{1,2} */
				  RETURN(_TIMEVAL); }
<AS>{D}+%			{ /* <AS>{D}{1,3}% */
				  RETURN(_TIMEVAL); }
<AS>\"[^\"]*\"			{ yytext[yyleng-1] = '\0';
				  cfglval.sval = mystrdup(yytext+1, _cfglex_);
				  return _STR;
				}
<ASE>\"[^\"]*\"			{ BEGIN(AS);
				  RETURN(_STR);
				}
<AS>{nWN}+			{ RETURN(_WORD); }

<PS,FS>{WN}*path{WND}*		{ return _PATH; }
<PS,FS>{WN}*{D}*{WN}*mode{WN}*	{ return _MODE; }
<PS,FS>{WN}*uid{WN}*		{ return _UID; }
<PS,FS>{WN}*gid{WN}*		{ return _GID; }
<PS,FS>{W}*{D}+{W}*		{ /* <PS,FS>{W}*{D}{1,5}{W}* */
				  /* trim leading & trailing spaces */
				  jj = yyleng-1;
				  while (isspace((int)*(yytext+jj))) {
					*(yytext+jj) = '\0';
					jj--;
				  }
				  jj = 0;
				  while (isspace((int)*(yytext+jj)))
					jj++;
				  /* if we've seen the 3rd _INT (for mode,
				     uid & gid), we're done with the filespec,
				     and we move on to the file body
				   */
				  if (++ii >= 3)
					BEGIN(LS);
				  cfglval.sval = mystrdup(yytext+jj, _cfglex_);
				  return _INT;
				}
<PS,FS>\"[^\"]*\"/{WND}+mode	{ /* {WND} in the pattern match, since we must
				     allow for the possibility of 'mode ...'
				     appearing in a subsequent line but after
				     an intervening lineno line; see also the
				     _PATH & _MODE pattern matches above
				   */
				  yytext[yyleng-1] = '\0';
				  cfglval.sval = mystrdup(yytext+1, _cfglex_);
				  return _STR;
				}

<PS,FS>.			{ unput(yytext[0]);
				  BEGIN(LS);
				}
<LS>\t\n			{ cfglval.sval = mystrdup(nullstring, _cfglex_);
				  return _LINE;
				}
<LS>{W}.+/\n			{ cfglval.sval = mystrdup(yytext+1, _cfglex_);
				  return _LINE;
				}

<OS>{nWN}[^\n]*			{ RETURN(_OBJECT); }

<*>","				{ return yytext[0]; }

<*>{W}				{ ; }
<*>\n				{ ; }
<*>\r				{ ; }

<*>.				{ return yytext[0]; }

%%

/* cfg error handler
 */
void
cfgerror (const char *message)
{

	if (stzerr)
		return;
	else {
		stzerr = TRUE;
		cfgerr = TRUE;
		if (cfgline && strcmp(cfgline,cfgbof[cfgcurr]) && (cfgcurr <= cfglast)) {
			if (cfgfileprev && strcmp(cfgfileprev, cfgfilecurr)) {
				logreturn(OUTPUT, LOG_ERR, ERRID_CFG_PARSE_ERROR, __FILE__, __LINE__, "cfgparse", FALSE, "%s (or possibly %s), at or near line %d: %s: %s", cfgfilecurr, cfgfileprev, lineno, message, yytext ? yytext : nullstring);
			} else {
				logreturn(OUTPUT, LOG_ERR, ERRID_CFG_PARSE_ERROR, __FILE__, __LINE__, "cfgparse", FALSE, "%s, at or near line %d: %s: %s", cfgfilecurr, lineno, message, yytext ? yytext : nullstring);
			}
		} else {
			if (cfgfileprev) {
				if (cfgfilecurr && strcmp(cfgfileprev, cfgfilecurr)) {
					logreturn(OUTPUT, LOG_ERR, ERRID_CFG_PARSE_ERROR, __FILE__, __LINE__, "cfgparse", FALSE, "%s (or possibly %s), at or near line %d: %s", cfgfileprev, cfgfilecurr, lineno, message);
				} else {
					logreturn(OUTPUT, LOG_ERR, ERRID_CFG_PARSE_ERROR, __FILE__, __LINE__, "cfgparse", FALSE, "%s, at or near line %d: %s", cfgfileprev, lineno, message);
				}
			} else /* if (cfgfilecurr) */ {
				logreturn(OUTPUT, LOG_ERR, ERRID_CFG_PARSE_ERROR, __FILE__, __LINE__, "cfgparse", FALSE, "%s, at or near line %d: %s", cfgfilecurr, lineno, message);
			}
		}
		bypass = TRUE;
		BEGIN(INITIAL);
		return;
	}
}

/* for config files macros.cfg, alarms.cfg, alerts.cfg, programs.cfg,
   files.cfg & objects.cfg (not systems.cfg or defines.cfg), feed their
   contents to the lexer input stream;
 */
signed int
cfginput (void)
{
	static char fnnam[] = "cfginput";
	static cfgln *clp;
	static char *line = (char *) NULL;
	static char *freeline = (char *) NULL;
	static char linenum[8];
	register char c;

	if ((cfgcurr > cfglast) /* || (cfgcurr > OBJCFG) */) {
		loffset = 0;
		return EOF;
	}

	c = ' ';	/* junk return value, just in case */

	if (! loffset) {
		if (bof) {
			cfgline = line = (char *) cfgbof[cfgcurr];
			clp = clpcurr = configs[cfgcurr];
			cfgfileprev = cfgfilecurr;
			cfgfilecurr = (char *) cfgfilenames[cfgcurr];
			/* zero this flag; at end of config lines, if this
			   is still zero, we know that no config lines were
			   processed for current host
			 */
			linenum[0] = '\0';
		} else if (bon) {
			if (! clp) {
				if (isverboselog)
					logreturn(PIKTLOG, LOG_WARNING, ERRID_CFG_FILE_MISSING, __FILE__, __LINE__, fnnam, FALSE, "missing %s", cfgfilecurr);
				cfgcurr++;
				if ((cfgcurr > cfglast) /* || (cfgcurr > OBJCFG) */) {
					loffset = 0;
					return EOF;
				} else {
					loffset = 0;
					lineno = 1;
					bof = TRUE;
					bon = FALSE;
					bol = FALSE;
					return '\177';	/* dummy char */
				}
			} else {
				while ((! bsmember(clp->bs, hbscurr, hbslongs)) || ! clp->doproc) {
					clp = clpcurr = clp->next;
					if (clp) {
						cfgfileprev = cfgfilecurr;
						cfgfilecurr = clp->file;
					} else {
						if (! linenum[0])
							logreturn(PIKTLOG, LOG_WARNING, ERRID_CFG_NO_INPUT, __FILE__, __LINE__, fnnam, FALSE, "%s, for host %s, no input!", cfgfileprev ? cfgfileprev : cfgfilecurr, hostcurr);
						cfgcurr++;
						if ((cfgcurr > cfglast) /* || (cfgcurr > OBJCFG) */) {
							loffset = 0;
							return EOF;
						} else {
							loffset = 0;
							lineno = 1;
							bof = TRUE;
							bon = FALSE;
							bol = FALSE;
							return '\177';	/* dummy char */
						}
					}
				}
			}
			sprintf(linenum, "%-6d\n", clp->lineno);
			cfgline = line = linenum;
		} else if (bol) {
			if (clp->domcr && domacros) {
				freeline = cfgline = line = expand(clp->line, lineno);
			} else {
				freeline = (char *) NULL;
				cfgline = line = clp->line;
			}
		}
	}

	c = *(line+loffset++);
	#ifdef DEEPDEBUG
	logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "YYSTATE is %d, YY_AT_BOL() is %d, c is %d %c", YYSTATE, YY_AT_BOL(), c, c);
	#endif

	if ((c == '\n') && ! *(line+loffset)) {
		if (bof) {
			bof = FALSE;
			bon = TRUE;
		} else if (bon) {
			bon = FALSE;
			bol = TRUE;
		} else if (bol) {
			FREE(freeline);
			if (clp->next) {
				bon = TRUE;
				clp = clpcurr = clp->next;
				if (clp) {
					cfgfileprev = cfgfilecurr;
					cfgfilecurr = clp->file;
				}
			} else {
				cfgcurr++;
				bof = TRUE;
				bon = FALSE;
				bol = FALSE;
			}
		}
		loffset = 0;
	}

	return c;

}

/* register a variable (STRING/NUMBER); note: although varnum[] & varstr[]
   should be plenty long, practically speaking, we do not now check for
   overflow
 */
int
regvar (char *name, int dattype)
{
	static char fnnam[] = "regvar";

	if (dattype == NUMBER) {
		strcat(varnum, name);
		strcat(varnum, " ");
	} else if (dattype == STRING) {
		strcat(varstr, name);
		strcat(varstr, " ");
	} else {
		logreturn(OUTPUT, LOG_ERR, ERRID_DATTYPE_INVALID, __FILE__, __LINE__, fnnam, FALSE, "invalid dat type: %s", name);
		bailout = TRUE;
		return ERROR;
	}
	return dattype;
}

/* return a var's dat type (STRING/NUMBER)
 */
int
vardattype (char *name, int namelen, cfgln *cfgline)
{
	static char fnnam[] = "vardattype";
	char *nam;
	cfgln *clp;
	static char *line = (char *) NULL;
	static char *freeline = (char *) NULL;
	char *cp;
	char c;
	int dqt = 0;
	int escape = FALSE;
	int typ;

	typ = ERROR;
	nam = (char *) mymalloc((strlen(name)+3) * sizeof(char), fnnam);
	sprintf(nam, " %s ", name);
	/* check to see if one of reserved var names */
	if (strstr(" inlin inline inputline outlin outline outputline rdlin rdline readline ", nam)) {
		typ = STRING;
	/* check to see if one of varstr's regvar()'ed already */
	} else if (strstr(varstr, nam)) {
		typ =  STRING;
	/* check to see if one of varnum's regvar()'ed already */
	} else if (strstr(varnum, nam)) {
		typ = NUMBER;
	}
	free(nam);
	if (typ != ERROR)
		return typ;
	/* look ahead in alarms.cfg to determine dat type */
	clp = cfgline;
	while (TRUE) {
		FREE(freeline);
		if (clp->domcr && domacros) {
			freeline = line = expand(clp->line, lineno);
		} else {
			freeline = (char *) NULL;
			line = clp->line;
		}
		/* look for name in line */
		cp = line;
		dqt = 0;
		escape = FALSE;
		while ((c = *cp)) {
			switch (c) {
				case '\\':
					if (! escape) {
						escape = TRUE;
					} else {
						escape = FALSE;
					}
					break;
				case '\"':
					if (! escape) {
						dqt++;
					}
					escape = FALSE;
					break;
				case '#':
				case '$':
					if (    ! escape
					     && even(dqt)
					     && *(cp+1)
					     && ! strncmp(name, cp+1, (unsigned int)namelen)
					     && ! (    isalnum((int)*(cp+1+namelen))
					            || (*(cp+1+namelen) == '_')
					          )
					   ) {
						nam = mystrdup(cp+1, fnnam);
						*(nam+namelen) = '\0';
						typ = regvar(nam, (c == '#') ? NUMBER : STRING);
						free(nam);
						if (typ != ERROR)
							return typ;
					}
					escape = FALSE;
					break;
				default:
					escape = FALSE;
					break;
			}
			cp++;
		}
		do {
			clp = clp->next;
		} while (clp && (isdigit((int)*clp->line) || ! bsmember(clp->bs, hbscurr, hbslongs) || ! clp->doproc));
		if (! clp || isalpha((int)*clp->line)) {
			break;
		}
	}
	return typ;
}
