/*
   piktc_svc.c -- PIKT master control program (piktc) service daemon

   $Header: /var/cvs/PIKT/pikt/src/piktc/piktc_svc.c,v 1.55 2007/01/02 09:08:48 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2006 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org

   --------------------------------------------------------------------------

   Usage:  piktc_svc [-GhTVz]

           -G              run in debug mode
           -h              show program help
           -T              run in test mode
           -V              show version info
           -z              run in 'make check' mode

 */

#include "../lib/common.h"
#include "piktc.h"

const char piktc_svc_version[] = "PIKT piktc_svc version " VERSION;

/* command-line options */
/* int ischeck = FALSE; */
int isdebug = FALSE;
/* int isdelete = FALSE; */
/* int isdoublecheck = FALSE; */
/* int iscksum = FALSE; */
/* int isdiff = FALSE; */
/* int isdisable = FALSE; */
/* int isenable = FALSE; */
int isexec = FALSE;
int isexecwait = FALSE;
/* int isexpand_macros = FALSE; */
int ishelp = FALSE;
/* int isinstall = FALSE; */
/* int isupdate_includes = FALSE; */
/* int iskill_piktd = FALSE; */
/* int iskill_piktc_svc = FALSE; */
/* int islist_files = FALSE; */
/* int islist_items = FALSE; */
/* int isrestart_piktd = FALSE; */
/* int isrestart_piktc_svc = FALSE; */
/* int isshow = FALSE; */
/* int isshow_status = FALSE; */
int isshow_version = FALSE;
int istest = FALSE;
/* int isverbose = FALSE; */
int ismakecheck = FALSE;

int bailout = FALSE;	/* signifies command-line error */

int lineno = 0;

const char nullstring[] = { '\0' };

const signed char err[]    = { ERROR, '\0' };
const signed char nil[]    = { NIL, '\0' };
const char errstr[] = "<<ERR>>";
const char nilstr[] = "<<NIL>>";

/* const signed char dquote[] = { DQTPROXY, '\0' }; */
/* const signed char squote[] = { SQTPROXY, '\0' }; */
/* const char space[]  = " "; */

char *hostserv;
char *rqstserv;
int statserv;

unsigned long services = ALL_SERVICES;	/* all services enabled by default */

unsigned int fork_delay = 1;

int
main (int argc, char *argv[])
{
	static char fnnam[] = "main";
	pid_t pid;
	static char pidstr[8];

	init(argc, argv, (void (*)(void))piktc_svc_usage, piktc_svc_version);
	if (bailout)
		exit(exstat);

	if (    ! isfile(filecmd, TRUE)
	     || ! isfile(pscmd, TRUE)
	     || ! isfile(arpcmd, TRUE)
	     || ! isfile(egrepcmd, TRUE)
	   ) {
		/* error(s) logged in PIKTLOG */
		exit(exstat);
	}

	/* check for blocking lockfile */
	if (islock(piktc_svclock)) {
		logexit(OUTPUT, LOG_ERR, ERRID_MULTIPLE_PIKTC_SVC, __FILE__, __LINE__, fnnam, FALSE, "%s found!  Another piktc_svc may be running.  Try killing the running piktc_svc (else rm the %s lockfile).", piktc_svclock, piktc_svclock);
	}
	/* check for running piktc_svc */
	if (procid("piktc_svc", PROCID) == ERROR) {
		logexit(OUTPUT, LOG_ERR, ERRID_MULTIPLE_PIKTC_SVC, __FILE__, __LINE__, fnnam, FALSE, "Another piktc_svc may be running.  Try killing the running piktc_svc (and/or rm the %s lockfile, if any).", piktc_svclock);
	}

	/* go into daemon mode */
	daemon_init(TRUE);

	/* determine the daemon pid */
	pid = getpid();
	/* when the piktc_svc main() calls piktc_svc_main(), possible
	   subsequent calls to fork() will render this pid invalid; what
	   to do? add some arbitrary guessed increment and hope the
	   ultimate pid comes out right? this caveat may only apply to
	   solaris
	 */
	/*
	pid += 1;
	 */
	sprintf(pidstr, "%u", (unsigned)pid);

	/* apply the lock, to prevent other concurrent daemon sessions */
	if (mklock(piktc_svclock, pidstr, TRUE, TRUE) == ERROR)
		logexit(PIKTLOG, LOG_ERR, ERRID_FILE_CREATE_ERROR, __FILE__, __LINE__, fnnam, FALSE, "mklock() error for %s", piktc_svclock);

	/* more initializations:  no initial service registrations */
	hostserv = mystrdup(nullstring, fnnam);
	rqstserv = mystrdup(nullstring, fnnam);

	logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "daemon (re)start");

	/* we pass control to the rpcgen'ed code at this point */

	piktc_svc_main();

	exit(exstat);	/* never reached */
}

/* output program usage to screen
 */
void
piktc_svc_usage (void)
{
	printf("\nUsage:  piktc_svc [-GhTVz]\n\n");
	printf("\t-G\t\t\trun in debug mode\n");
	printf("\t-h\t\t\tshow program help\n");
	printf("\t-T\t\t\trun in test mode\n");
	printf("\t-V\t\t\tshow version info\n\n");
	printf("\t-z\t\t\trun in 'make check' mode\n");
	return;
}
