%{

/*
   tag.y -- PIKT piktc tag parser:  deals with machine conditionals and the
            tagging of config file lines; tagging does not apply to
            systems.cfg!

   $Header: /var/cvs/PIKT/pikt/src/piktc/tag.y,v 1.30 2006/11/04 09:59:36 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2006 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
 */

#include "../lib/common.h"
#include "piktc.h"

char _tagparse_[] = "tagparse";

#define CHKIFNUM	if (ifnum > MAXIFS) { \
				logreturn(OUTPUT, LOG_ERR, ERRID_IFHOSTNUM_MAX_EXCEEDED, __FILE__, __LINE__, _tagparse_, FALSE, "%s, line %d, MAXIFS (%d) exceeded", cfgfilecurr, lineno-1, MAXIFS); \
				bailout = TRUE; \
			} else if (ifnum < 0) { \
				logexit(OUTPUT, LOG_ERR, ERRID_IFHOSTNUM_LESS_THAN_ZERO, __FILE__, __LINE__, _tagparse_, FALSE, "%s, line %d, ifnum < 0", cfgfilecurr, lineno-1); \
				bailout = TRUE; \
			}

#define SETBS(a,b)	tbs1 = a; \
			a = b; \
			FREEVOID(tbs1)

int ifnum;

unsigned long *hbscur[MAXIFS+1], *hbscum[MAXIFS+1];
unsigned long *tbs1, *tbs2, *tbs3;

int i;
char c;

%}

%union {
	char *sval;
	int ival;
	unsigned long *hbsval;
}

%token <sval> _IDENT _LINE _DEF
%token _IF _ELIF _ELSE _ENDIF
%type <hbsval> hset

%left '|'
%left '&'
%nonassoc '!'

%%

start:		{
			if (scanpass == 2) {
				ifnum = 0;
				hbscur[ifnum] = (unsigned long *) mycalloc(hbslongs, sizeof(unsigned long), _tagparse_);
				hbscum[ifnum] = (unsigned long *) mycalloc(hbslongs, sizeof(unsigned long), _tagparse_);
				fill(hbscur[ifnum], hbslongs);
			}
		} lines {
			if (scanpass == 2) {
				FREEVOID(hbscur[ifnum]);
				FREEVOID(hbscum[ifnum]);
			}
		}
		
		;

lines:		line
		| lines line
		;

line:		_LINE {
			if (scanpass == 1) {
				if (cfgcurr == DEFCFG)
					defnum++;
			} else if (scanpass == 2) {
				#ifdef DEEPDEBUG
				if (isdebug)
					logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, _tagparse_, FALSE, "%s", $1);
				#endif
				if (! islist_elements) {
					tag(lineno-1, cfgfilecurr, hbscur[ifnum], hbslongs, hasmacro ? domacros : FALSE, TRUE, $1);
				} else {
					if ((cfgcurr == SYSCFG) || (cfgcurr == DEFCFG)) {
						tag(lineno-1, cfgfilecurr, hbscur[ifnum], hbslongs, FALSE, TRUE, $1);
					} else {
						if (! isspace((int)*$1)) {
							if (*$1 == '#') {
								tag(lineno-1, cfgfilecurr, hbscur[ifnum], hbslongs, FALSE, TRUE, $1);
							} else {
								if (cfgcurr != MCRCFG) {
									i = 0;
									while ((c = *($1+i)) && (! isspace((int)c))) i++;
									if (c) {
										*($1+i) = '\n';
										*($1+i+1) = '\0';
									}
								}
								tag(lineno-1, cfgfilecurr, hbscur[ifnum], hbslongs, FALSE, TRUE, $1);
								switch (cfgcurr) {
									case MCRCFG:
										tag(lineno-1, cfgfilecurr, hbscur[ifnum], hbslongs, FALSE, TRUE, mystrdup("\t0\n", _tagparse_));
										break;
									case ALMCFG:
										tag(lineno-1, cfgfilecurr, hbscur[ifnum], hbslongs, FALSE, TRUE, mystrdup("\tquit\n", _tagparse_));
										break;
									case ALTCFG:
										tag(lineno-1, cfgfilecurr, hbscur[ifnum], hbslongs, FALSE, TRUE, mystrdup("\texeccmd\n\t\"echo\"\n", _tagparse_));
										break;
									case PRGCFG:
									case FILCFG:
										tag(lineno-1, cfgfilecurr, hbscur[ifnum], hbslongs, FALSE, TRUE, mystrdup("\tp\ath \"/foo\" mode 000 uid 0 gid 0\n\t0\n", _tagparse_));
										break;
									case OBJCFG:
										tag(lineno-1, cfgfilecurr, hbscur[ifnum], hbslongs, FALSE, TRUE, mystrdup("\t0\n", _tagparse_));
										break;
								}
							}
						} else {
							free($1);
						}
					}
				}
			}
		}
		| _DEF {
			if (scanpass == 2)
				tag(lineno-1, cfgfilecurr, hbscur[ifnum], hbslongs, hasmacro ? domacros : FALSE, FALSE, $1);
		}
		| conditional
		;

conditional:	if_ _ENDIF {
			if (scanpass == 2) {
				FREEVOID(hbscur[ifnum]);
				FREEVOID(hbscum[ifnum]);
				ifnum--;
				CHKIFNUM;
			}
		}
		| if_ else_ _ENDIF {
			if (scanpass == 2) {
				FREEVOID(hbscur[ifnum]);
				FREEVOID(hbscum[ifnum]);
				ifnum--;
				CHKIFNUM;
			}
		}
		| if_ elifs_ _ENDIF {
			if (scanpass == 2) {
				FREEVOID(hbscur[ifnum]);
				FREEVOID(hbscum[ifnum]);
				ifnum--;
				CHKIFNUM;
			}
		}
		| if_ elifs_ else_ _ENDIF {
			if (scanpass == 2) {
				FREEVOID(hbscur[ifnum]);
				FREEVOID(hbscum[ifnum]);
				ifnum--;
				CHKIFNUM;
			}
		}
		;

if_:		if
		| if lines
		;

if:		_IF hset {
			if (scanpass == 2) {
				ifnum++;
				CHKIFNUM;
				hbscur[ifnum] = bsintersection($2, hbscur[ifnum-1], hbslongs);
				hbscum[ifnum] = bsintersection($2, hbscur[ifnum-1], hbslongs);
				free($2);
			}
		}
		;

elifs_:		elif
		| elif lines
		| elifs_ elif
		| elifs_ elif lines
		;

elif:		_ELIF hset {
			if (scanpass == 2) {
				tbs2 = bscomplement(hbscum[ifnum], hbslongs);
				tbs3 = bsintersection($2, tbs2, hbslongs);
				free(tbs2);
				SETBS(hbscur[ifnum], bsintersection(tbs3, hbscur[ifnum-1], hbslongs));
				free(tbs3);
				tbs2 = bsunion($2, hbscum[ifnum], hbslongs);
				SETBS(hbscum[ifnum], bsintersection(tbs2, hbscur[ifnum-1], hbslongs));
				free(tbs2);
				free($2);
			}
		}
		;

else_:		else
		| else lines
		;

else:		_ELSE {
			if (scanpass == 2) {
				tbs2 = bscomplement(hbscum[ifnum], hbslongs);
				SETBS(hbscur[ifnum], bsintersection(tbs2, hbscur[ifnum-1], hbslongs));
				free(tbs2);
			}
		}
		;

hset:		_IDENT {
			if (scanpass == 2) {
				/* HST is a stand-in for OPS, GRP & ALS */
				$$ = bsdup(getbs($1, HST, TRUE), hbslongs);
				free($1);
			}
		}
		| '(' hset ')' {
			if (scanpass == 2) {
				$$ = $2;
				/* or :
				$$ = bsdup($2, hbslongs);
				free($2);
				 */
			}
		}
		| '!' hset {
			if (scanpass == 2) {
				$$ = bscomplement($2, hbslongs);
				free($2);
			}
		}
		| hset '&' hset {
			if (scanpass == 2) {
				$$ = bsintersection($1, $3, hbslongs);
				free($1);
				free($3);
			}
		}
		| hset '|' hset {
			if (scanpass == 2) {
				$$ = bsunion($1, $3, hbslongs);
				free($1);
				free($3);
			}
		}
		;

%%
