/*
   regexp.c -- PIKT regular expression routines

   $Header: /var/cvs/PIKT/pikt/src/piktd/regexp.c,v 1.31 2006/11/04 09:59:36 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2006 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
*/

#include "../lib/common.h"
#include "piktd.h"

regmatch_t regmatch[MAXVALSCN+1];

/* when invoke #split(), we inspect its second parameter (if any) to
   determine if this calls for a pattern match; if not, simply return with
   the value NIL indicating that a regexp pattern match does not apply;
   otherwise, this is a legitimate call for a regular expression pattern
   match, so pass the given string and regexp along to the GNU rx routines;
   if we have parenthesized patterns, and we have a match, copy the matched
   substrings to the valscan[] array as necessary; set matches equal to
   the number of such matches; return TRUE or FALSE, depending on whether
   or not the given string pattern matches the regexp
 */
int
regexp (int vrndx, char *str, char *exp, int cflags, int isparsefn)
{
	static char fnnam[] = "regexp";
	regex_t preg;
	int regerrcode;
	char regerrmsg[MAXLINE+1];
	register int i, k;
	register char c;
	int islp;
	int openlp;
	int nmatch;
	int len;
	vl *vlp;

	vlp = (vl *) NULL;
	len = 0;
	freevalscn();
	islp = FALSE;
	openlp = FALSE;
	nmatch = 1;
	i = 0;
	while ((c = *(exp+i))) {
		i++;
		switch (c) {
			case '\\':
				islp = FALSE;
				if (*(exp+i))
					i++;
				break;
			case '(':
				islp = TRUE;
				openlp = TRUE;
				break;
			case ')':
				if (openlp && ! islp) {
					nmatch++;
					openlp = FALSE;
				}
				/* no break here! */
			default:
				islp = FALSE;
				break;
		}
	}
	if (isparsefn && (nmatch == 1))
		return NIL;
	if ((regerrcode = regcomp(&preg, exp, cflags))) {
		regerror(regerrcode, &preg, regerrmsg, MAXLINE);
		logreturn(PIKTLOG, LOG_ERR, ERRID_REGCOMP_ERROR, __FILE__, __LINE__, fnnam, FALSE, "in %s, regcomp() error for regexp %s: %s", ALMNAM, exp, regerrmsg);
		if (doeval) exprres = ERROR;
		return FALSE;
	}
	regerrcode = regexec(&preg, str, (size_t) nmatch, regmatch, 0 /* REG_NOTBOL|REG_NOTEOL */);
	regfree(&preg);
	if (regerrcode) {
		matches = 0;
		return FALSE;
	} else {
		matches = nmatch-1;
		for (i=0; i<=matches; i++) {
			len = regmatch[i].rm_eo - regmatch[i].rm_so;
			valscn[i] = (char *) mymalloc((size_t)(len+1), fnnam);
			strncpy(valscn[i], str+regmatch[i].rm_so, (size_t)len);
			*(valscn[i]+len) = '\0';
			if (vrndx) {
				for (k=0; k<=MAXARYDIM; k++)
					arndx[k] = 0L;
				arndx[1] = (signed long) i;
				if ((vlp = lookupvl(valnow, vrndx, arndx, (char *)NULL, ARYTYP(vrndx)))) {
					FREE(vlp->str);
				} else {
					if (! (vlp = regvl(valnow, vrndx, arndx, (char *)NULL, ARYTYP(vrndx)))) {
						logreturn(PIKTLOG, LOG_ERR, ERRID_REGVL_ERROR, __FILE__, __LINE__, fnnam, FALSE, "in %s, regvl() error for var %s", ALMNAM, VARNAM(vrndx));
						matches = 0;
						return FALSE;
					}
				}
				vlp->str = mystrdup(valscn[i], fnnam);
				errno = 0;
				vlp->num = (double) atof(vlp->str);
				if (errno)
					vlp->num = ERROR;
			}
			if ((i == 0) && (POSTYP == REGEXPR)) {
				if (*(valscn[0]+len-1) == '\n')
					*(valscn[0]+len-1) = '\0';
			}
		}
		/* when parsing the input line using regexp, chop the
		   linefeed, if present, from the last field (this was
		   done for valscn[0] above)
		 */
		if (POSTYP == REGEXPR) {
			if (*(valscn[i-1]+len-1) == '\n') {
				*(valscn[i-1]+len-1) = '\0';
				if (vrndx && vlp)
					*((vlp->str)+len-1) = '\0';
			}
		}
		return TRUE;
	}
}
